import os
from ..imports import *

def get_http(domain):
    
    http=None
    if domain.startswith('http'):
        http = 'http'
        if domain.startswith('https'):
            http = f"{http}s"
    return http
def strip_http(domain,http=None):
    http = http or get_http(domain)
   
    domain = domain[len(http or ''):]
    
    domain = eatAll(domain,[':','/'])
    return domain

    
def get_www(domain,http=None):
    http = http or get_http(domain)
    domain = strip_http(domain = domain,http=http)
    if domain.startswith('www.'):
        return True
    return False   
def strip_www(domain,http=None):
    http = http or get_http(domain)
    
    domain = strip_http(domain = domain,http=http)
    
    www = get_www(domain,http=None)
    if www:
        domain = domain[len('www.'):]
    if http:
        
        domain = f"{http}://{domain}"
    return domain

def get_http_www(domain):
    http = get_http(domain)
    www = get_www(domain,http=http)
    return {"http":http,"www":www}
def strip_http_www(domain,http=None):
    domain = strip_www(domain,http=http)
    domain = strip_http(domain,http=http)
    return domain
def get_stripped_domain(domain,http=None):
    url,params = get_url_params(domain,get_url=True)
    url = strip_http_www(url,http=http)
    return url.split('/')
def get_url_params(url,get_url=False):
    params = {}
    if '?' in url:
        url,nuparams = url.split('?')
        nuparams = nuparams.split('&')
        for param in nuparams:
            param_spl = param.split('=')
            key = param_spl[0]
            params[key]=None
            if len(param_spl) >1:
               params[key] = '='.join(param_spl[1:])
    if get_url:
        return url,params
    return params
def get_cleaned_url(url):
    url,params = get_url_params(url,get_url=True)
    return url
def get_domain_paths(domain,http=None):
    domain = strip_http_www(domain,http=http)
    domain_paths = domain.split('/')
    paths = domain_paths[1:] if len(domain_paths)>1 else []
    return {"domain":domain_paths[0],"path":'/'.join(paths)}
def get_domain(domain,http=None):
    domain_ext = get_domain_paths(domain,http=http)
    return domain_ext.get("domain")
def get_domain_path(domain,http=None):
    domain_ext = get_domain_paths(domain,http=http)
    return domain_ext.get("path")
def get_domain_name_ext(domain,http=None):
    domain = get_domain(domain,http=http)
    domain_ext = os.path.splitext(domain)
    return {"name":domain_ext[0],"ext":domain_ext[-1]}
def get_extension(domain=None,http=None,parsed=None,options=None):
    if parsed:
        try:
            domain = parsed.netloc
            http = parsed.scheme
        except:
            pass
    domain_ext = get_domain_name_ext(domain,http=http)
    return domain_ext.get("ext")
def get_domain_name(domain,http=None):
    domain_ext = get_domain_name_ext(domain,http=http)
    return domain_ext.get("name")
def get_url_js(url):
    www = get_http_www(url)
    http = get_http(url)
    domain = get_domain(url)
    stripped_domain = get_stripped_domain(url,http=http)
    params = get_url_params(url)
    domain_name_ext = get_domain_name_ext(url,http=http)
    domain_paths = get_domain_paths(url,http=http)
    cleaned_url = get_cleaned_url(url)
    return {"www":www,"http":http,"domain":domain,"cleaned_url":cleaned_url,"stripped_domain":stripped_domain,"params":params,"domain_name_ext":domain_name_ext,"domain_paths":domain_paths}

get_extention = get_extension




PINTEREST_HOSTS = {"pinterest.com", "www.pinterest.com"}

def detect_pinterest(parsed):
    if parsed.netloc.lower() not in PINTEREST_HOSTS:
        return None

    parts = parsed.path.split("/")
    if len(parts) > 2 and parts[1] == "pin":
        return {"downloadable": True, "kind": "image", "provider": "pinterest",
                "id": parts[2], "direct": False}

    return {"downloadable": False, "provider": "pinterest"}



from ..imports import *
from ..urlManager import *
from ..requestManager import *
from ..soupManager import *
import json,subprocess,requests
from pathlib import Path
from urllib.parse import urlparse



from .src import *
seleniumManager = seleneumManager
def normalize_url(url, base_url=None):
    manager = seleniumManager(url)
    base_url = manager.base_url
    if url.startswith(base_url):
        url = url[len(base_url):]
    normalized_url = urljoin(base_url, url.split('#')[0])
    if not normalized_url.startswith(base_url):
        return None
    return normalized_url



# abstract_youtube/downloader.py

from .imports import *
class YouTubeDownloadError(Exception):
    pass


class AbstractYouTubeDownloader:
    """
    Responsible ONLY for:
    - managing session
    - validating googlevideo URLs
    - downloading media

    It does NOT:
    - scrape HTML
    - decipher signatures
    - choose formats
    """

    def __init__(self, *, user_agent: str | None = None):
        self.session = requests.Session()
        self.session.headers.update({
            "User-Agent": user_agent or (
                "Mozilla/5.0 (X11; Linux x86_64) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/120 Safari/537.36"
            ),
            "Referer": "https://www.youtube.com/",
            "Origin": "https://www.youtube.com",
            "Accept": "*/*",
            "Accept-Encoding": "identity",
            "Connection": "keep-alive",
        })

    # ---------------------------
    # VALIDATION
    # ---------------------------

    @staticmethod
    def validate_direct_url(url: str) -> None:
        p = urlparse(url)
        q = parse_qs(p.query)

        if "googlevideo.com" not in p.netloc:
            raise YouTubeDownloadError("Not a googlevideo URL")

        for key in ("itag", "mime", "expire"):
            if key not in q:
                raise YouTubeDownloadError(f"Missing required param: {key}")

        if "sig" not in q and "signature" not in q:
            raise YouTubeDownloadError("Missing signature")

        expire = int(q["expire"][0])
        if expire < time.time():
            raise YouTubeDownloadError("URL expired")

    # ---------------------------
    # DOWNLOAD
    # ---------------------------

    def download(
        self,
        *,
        url: str,
        output_path: str,
        chunk_size: int = 1024 * 1024,
    ) -> str:
        """
        Downloads a VALID googlevideo URL.
        """
        self.validate_direct_url(url)

        with self.session.get(url, stream=True, timeout=30) as r:
            r.raise_for_status()
            with open(output_path, "wb") as f:
                for chunk in r.iter_content(chunk_size):
                    if chunk:
                        f.write(chunk)

        return output_path



from __future__ import annotations
from abstract_utilities import get_home_dir
from .imports import *
import os,tempfile,shutil,requests
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Dict, Any
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.options import Options
logger = get_logFile(__name__)
def make_writable_runtime_dir(base: str = None) -> Path:
    tmp_path = "tmp/selenium-runtime"
    base = base or os.path.join(get_home_dir(),tmp_path)
    path = Path(base)
    path.mkdir(parents=True, exist_ok=True)
    path.chmod(0o700)
    return path
@contextmanager
def chrome_driver(
    headless: bool = True,
    referer: str | None = None,
    user_agent: str | None = None,
    headers: Optional[dict[str, str]] = None,
):
    runtime_root = make_writable_runtime_dir()

    profile_dir = tempfile.mkdtemp(prefix="chrome-profile-", dir=str(runtime_root))
    cache_dir = tempfile.mkdtemp(prefix="chrome-cache-", dir=str(runtime_root))

    options = Options()

    if headless:
        options.add_argument("--headless=new")

    options.add_argument(f"--user-data-dir={profile_dir}")
    options.add_argument(f"--disk-cache-dir={cache_dir}")

    if user_agent:
        options.add_argument(f"--user-agent={user_agent}")

    options.add_argument("--no-sandbox")
    options.add_argument("--disable-dev-shm-usage")
    options.add_argument("--disable-gpu")
    options.add_argument("--disable-extensions")
    options.add_argument("--disable-background-networking")
    options.add_argument("--disable-sync")
    options.add_argument("--metrics-recording-only")
    options.add_argument("--no-first-run")
    options.add_argument("--no-default-browser-check")

    driver = None

    try:
        driver = webdriver.Chrome(options=options)

        if headers:
            driver.execute_cdp_cmd("Network.enable", {})

            extra_headers = {
                key: value
                for key, value in headers.items()
                if key.lower()
                not in {"user-agent", "host", "connection", "content-length"}
            }

            if referer and "Referer" not in extra_headers:
                extra_headers["Referer"] = referer

            if extra_headers:
                driver.execute_cdp_cmd(
                    "Network.setExtraHTTPHeaders",
                    {"headers": extra_headers},
                )

        yield driver

    finally:
        if driver is not None:
            driver.quit()

        shutil.rmtree(profile_dir, ignore_errors=True)
        shutil.rmtree(cache_dir, ignore_errors=True)




def get_rendered_html(
    url: str,
    timeout: int = 30,
    referer: str | None = None,
    user_agent: str | None = None,
    headers: Optional[dict[str, str]] = None,
) -> str:
    with chrome_driver(
        headless=True,
        referer=referer,
        user_agent=user_agent,
        headers=headers,
    ) as driver:
        driver.set_page_load_timeout(timeout)
        driver.get(url)
        return driver.page_source
def get_requested_html(
    url: str,
    session: requests.Session | None = None,
    headers: Optional[dict[str, str]] = None,
    timeout: int = 30,
    referer: str | None = None,
    user_agent: str | None = None,
) -> str:
    owns_session = session is None
    session = session or requests.Session()

    request_headers = dict(headers or {})

    if user_agent and "User-Agent" not in request_headers:
        request_headers["User-Agent"] = user_agent

    if referer and "Referer" not in request_headers:
        request_headers["Referer"] = referer

    try:
        with session.get(
            url,
            headers=request_headers or None,
            timeout=timeout,
        ) as response:
            response.raise_for_status()
            return response.text

    finally:
        if owns_session:
            session.close()

def fetch_html(
    url: str,
    session: requests.Session | None = None,
    headers: Optional[dict[str, str]] = None,
    timeout: int = 30,
    referer: str | None = None,
    user_agent: str | None = None,
) -> str:
    try:
        return get_rendered_html(
            url=url,
            timeout=timeout,
            referer=referer,
            user_agent=user_agent,
            headers=headers,
        )

    except Exception as exc:
        print(f"Selenium failed for {url}, falling back to requests: {exc}")

        return get_requested_html(
            url=url,
            session=session,
            headers=headers,
            timeout=timeout,
            referer=referer,
            user_agent=user_agent,
        )



from .imports import *
from .functions import *

class TLSAdapter(HTTPAdapter):
    """
    Requests adapter that injects a preconfigured SSLContext (from SSLManager)
    into both the main pool and any HTTPS proxy pools.
    """
    def __init__(
        self,
        ssl_manager: Optional[SSLManager] = None,
        ciphers: Optional[Union[str, Sequence[str]]] = None,
        certification: Optional[int] = None,   # e.g., ssl.CERT_REQUIRED
        ssl_options: Optional[int] = None
    ) -> None:
        ciphers_str = normalize_ciphers(ciphers)

        self.ssl_manager = ssl_manager or SSLManager(
            ciphers=ciphers_str,
            ssl_options=ssl_options,
            certification=certification,
        )
        # Expose normalized/canonical values for singleton comparisons
        self.ciphers: Optional[str] = getattr(self.ssl_manager, "ciphers", ciphers_str)
        self.certification: int = getattr(self.ssl_manager, "certification", ssl.CERT_REQUIRED)
        self.ssl_options: Optional[int] = getattr(self.ssl_manager, "ssl_options", None)
        self.ssl_context: ssl.SSLContext = self.ssl_manager.ssl_context

        super().__init__()

    # Use canonical signature for clarity across urllib3 versions
    def init_poolmanager(self, connections: int, maxsize: int, block: bool = False, **pool_kwargs) -> None:
        pool_kwargs["ssl_context"] = self.ssl_context
        super().init_poolmanager(connections, maxsize, block=block, **pool_kwargs)

    def proxy_manager_for(self, proxy: str, **proxy_kwargs) -> ProxyManager:
        # only attach context for HTTPS proxies
        if proxy and str(proxy).lower().startswith("https://"):
            proxy_kwargs["ssl_context"] = self.ssl_context
        return super().proxy_manager_for(proxy, **proxy_kwargs)






from __future__ import annotations
from ..imports import *
from ..functions import *
from dataclasses import asdict as _asdict





@dataclass
class ParsedURL:
    scheme:   str  = "https"
    host:     str  = ""
    port:     int | None = None
    www:      bool = False
    name:     str  = ""
    ext:      str  = ""
    path:     str  = ""
    query:    dict = field(default_factory=dict)
    fragment: str  = ""
    bare:     bool = False

    @property
    def netloc(self) -> str:
        return f"{self.host}:{self.port}" if self.port else self.host

    @property
    def full_domain(self) -> str:
        return f"{self.scheme}://{self.netloc}"

    @property
    def full_url(self) -> str:
        path = f"/{self.path.lstrip('/')}" if self.path else ""
        qs   = f"?{urlencode(self.query)}" if self.query else ""
        frag = f"#{self.fragment}" if self.fragment else ""
        return f"{self.full_domain}{path}{qs}{frag}"

    # ── dict protocol ────────────────────────────────────────────
    def _as_dict(self) -> dict:
        d = _asdict(self)
        d["netloc"]      = self.netloc
        d["full_domain"] = self.full_domain
        d["full_url"]    = self.full_url
        d["valid_variants"] = []   # bare hosts skip variant probing
        return d

    def get(self, key, default=None):
        return self._as_dict().get(key, default)

    def __getitem__(self, key):
        return self._as_dict()[key]

    def __contains__(self, key):
        return key in self._as_dict()



def _is_bare_host(host: str) -> bool:
    """True for localhost, IPs, and hostnames that have no public TLD."""
    if not host:
        return False
    bare = host.split(":")[0]
    if bare == "localhost":
        return True
    try:
        ipaddress.ip_address(bare)
        return True
    except ValueError:
        pass
    # no dot == no TLD
    return "." not in bare


def _parse_query(raw: str) -> dict:
    if not raw:
        return {}
    result = {}
    for pair in raw.split("&"):
        if not pair:
            continue
        k, _, v = pair.partition("=")
        result[k] = v
    return result


def parse_url_clean(url: str | None) -> ParsedURL:
    """
    Parse a URL into a fully explicit ParsedURL.
    Never produces None in string fields.
    Works for: https://example.com/path, http://localhost:8384/, 192.168.1.1:5000
    """
    if not url:
        return ParsedURL()

    url = url.strip()

    # inject scheme if missing so urlparse doesn't mangle netloc into path
    if "://" not in url:
        url = f"https://{url}"

    p: ParseResult = urlparse(url)

    scheme = p.scheme or "https"
    raw_host = p.hostname or ""        # always lowercase, no port
    port = p.port                      # int or None
    path = p.path.strip("/")
    query = _parse_query(p.query)
    fragment = p.fragment or ""

    www = raw_host.startswith("www.")
    host_no_www = raw_host.removeprefix("www.")
    bare = _is_bare_host(host_no_www)

    if bare:
        name = host_no_www.split(":")[0]
        ext  = ""
    else:
        name, ext = os.path.splitext(host_no_www)
        if not ext and "." in host_no_www:
            # splitext only fires on last component — fallback
            parts = host_no_www.rsplit(".", 1)
            name, ext = parts[0], f".{parts[1]}"

    return ParsedURL(
        scheme=scheme,
        host=raw_host,
        port=port,
        www=www,
        name=name,
        ext=ext,
        path=path,
        query=query,
        fragment=fragment,
        bare=bare,
    )
class urlManager:
    """
    urlManager for managing and cleaning URLs.

    Uses shared functions from domain_utils/url_utils/specUrl_utils.
    """

    def __init__(self, url=None, session=None, valid_variants=None):
        self.valid_variants = valid_variants or False
        self.session = session or requests.Session()
        self._url = url

        self.parsed = None
        if url is None:
            self.url = None
            self.clean_urls = []
            self.protocol = None
            self.domain = None
            self.path = ""
            self.query = {}
            self.all_urls = []
        else:
            self.update_url(url)

    def update_url(self, url):
        """Update the URL and refresh related attributes."""
        self._url = url
        if self._url == None:
            self.parsed = {}
        else:
            if self.valid_variants:
                self.parsed = parse_url(self._url,valid_variants=self.valid_variants)
            else:
                self.parsed = parse_url_clean(self._url)
        if url is None:
            self.clean_urls = self.parsed.get('valid_variants')
            self.url = self.parsed.get('full_url')
            self.protocol = None
            self.domain = None
            self.path = ""
            self.query = {}
            self.all_urls = []
            return

        self.valid_variants = self.parsed.get('valid_variants')
        self.path = self.parsed.get('path')
        self.clean_urls = [f"{variant}/{self.path}" for variant in self.valid_variants if variant]
        self.url = self.parsed.get('full_url')
        self.protocol = self.parsed.get("scheme")
        self.domain = self.parsed.get('full_domain')
        self.query = self.parsed.get("query")
        self.all_urls = self.clean_urls

    def _generate_variants(self, parsed) -> list:
        """Generate candidate URLs using ALL_URL_KEYS (scheme, www, extensions)."""
        netloc_data = parsed.get("netloc")
        base_domain = parsed.get("domain")
        ext = parsed.get("ext")
        variants = []

        for scheme in ALL_URL_KEYS["scheme"]:
            for www in ALL_URL_KEYS["netloc"]["www"]:
                for ext_cand in (ALL_URL_KEYS["netloc"]["extentions"][0] + ALL_URL_KEYS["netloc"]["extentions"][1]):
                    # skip if extension already fixed and doesn't match
                    if ext and ext != ext_cand:
                        continue
                    candidate_netloc = f"{'www.' if www else ''}{base_domain}{ext_cand or ext}"
                    candidate_url = f"{scheme}://{candidate_netloc}{parsed.get('path','')}"
                    variants.append(candidate_url)

        # Deduplicate
        seen = set()
        unique_variants = [v for v in variants if not (v in seen or seen.add(v))]
        return sorted(unique_variants, key=lambda v: (not v.startswith("https"), v))

    def get_correct_url(self, candidates=None) -> str:
        """
        Try candidates (HEAD request). 
        Return first that resolves with 200.
        """
        candidates = candidates or self.clean_urls
        for candidate in candidates:
            try:
                response = self.session.head(candidate, timeout=5, allow_redirects=True)
                if response.status_code == 200:
                    return candidate
            except requests.exceptions.RequestException as e:
                logging.info(f"Failed: {candidate} ({e})")
        return None

    # === convenience wrappers ===
    def get_domain(self) -> str:
        if not self.parsed:
            return None
        return reconstructNetLoc(self.parsed.get("netloc"))

    def is_valid_url(self, url=None):
        """Check if a URL parses into scheme+netloc."""
        url = url or self.url
        if not url:
            return False
        p = parse_url(url)
        domain=None
        scheme=None
        netloc=None
        if isinstance(p,dict):
            scheme = p.get("scheme")
            netloc = p.get("netloc", {})
            if isinstance(netloc,dict):
                domain = netloc.get("domain")
        return bool(scheme) and bool(domain or netloc)

    def make_valid(self, href, base_url=None):
        """
        Fix relative links by joining with a base URL.
        If base_url is not provided, uses self.url.
        """
        base_url = base_url or self.url
        if not base_url:
            return None
        return requests.compat.urljoin(base_url, href)


    def base_url(self):
        """Return base of current url (scheme+domain)."""
        if not self.parsed:
            return None
        scheme = self.parsed.get("scheme")
        netloc = reconstructNetLoc(self.parsed.get("netloc"))
        return f"{scheme}://{netloc}/"

    def url_basename(self):
        if not self.parsed:
            return ""
        return self.parsed.get("path", "").strip("/").split("/")[-1]
def get_url_mgr(url=None,url_mgr=None):
    return url_mgr or urlManager(url)
def get_url(url=None,url_mgr=None):
    url_mgr = get_url_mgr(url=url,url_mgr=url_mgr)
    return url_mgr.url




from abstract_apis import postRequest
from abstract_utilities import *
from abstract_math import divide_it,add_it,multiply_it,subtract_it
##logging.basicConfig(level=logging.INFO)
##logger  = get_logFile(__name__)

LOG_FORMAT = (
    "[%(asctime)s] "
    "%(levelname)-8s "
    "%(name)s:%(lineno)d | "
    "%(message)s"
)

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"




def get_logger(
    name: str,
    log_dir: str | Path = "logs",
    level: int = logging.INFO,
    console: bool = True,
    max_bytes: int = 5 * 1024 * 1024,
    backup_count: int = 5,
):
    logger = logging.getLogger(name)

    # Prevent duplicate handlers (VERY important)
    if logger.handlers:
        return logger

    logger.setLevel(level)

    formatter = logging.Formatter(LOG_FORMAT, DATE_FORMAT)

    log_dir = Path(log_dir)
    log_dir.mkdir(parents=True, exist_ok=True)

    # File handler (rotating)
    file_handler = RotatingFileHandler(
        log_dir / f"{name}.log",
        maxBytes=max_bytes,
        backupCount=backup_count,
        encoding="utf-8",
    )
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    # Console handler
    if console:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    logger.propagate = False
    return logger
logger = get_logFile("video_pipeline")



from contextlib import contextmanager
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.options import Options

CHROMIUM_BIN = '/usr/bin/chromium-browser'
CHROMEDRIVER_BIN = '/usr/local/bin/chromedriver'

def _build_options() -> Options:
    opts = Options()
    opts.binary_location = CHROMIUM_BIN
    opts.add_argument('--headless=new')
    opts.add_argument('--no-sandbox')
    opts.add_argument('--disable-dev-shm-usage')
    opts.add_argument('--disable-gpu')
    return opts

@contextmanager
def chrome_driver():
    service = Service(CHROMEDRIVER_BIN)
    driver = webdriver.Chrome(service=service, options=_build_options())
    try:
        yield driver
    finally:
        driver.quit()  # sends /shutdown to chromedriver, then SIGTERM to chrome

def get_url_source(url: str) -> str:
    with chrome_driver() as driver:
        driver.get(url)
        return driver.page_source



from .imports import *
from .paths import *
from .urls import *
def get_save_page_path(url, output_dir):
    """
    Save HTML page to local directory.
    """
    parsed_url = urlparse(url)
    page_path = parsed_url.path.lstrip('/')

    if not page_path or page_path.endswith('/'):
        page_path = os.path.join(page_path, 'index.html')
    elif not os.path.splitext(page_path)[1]:
        page_path += '.html'

    page_full_path = os.path.join(output_dir, page_path)
    return page_full_path
def save_page(url, content,output_dir):
    page_full_path = get_save_page_path(url=url,
                                        output_dir=output_dir)
    page_full_path = currate_full_path(page_full_path)
    if page_full_path:
        dirname = os.path.dirname(page_full_path)
        

        with open(page_full_path, 'w', encoding='utf-8') as f:
            f.write(content)
        print(f"Saved page: {page_full_path}")
def get_asset_path(asset_url,
                   base_url,
                   output_dir,
                   downloaded_assets=None,
                   session=None):
    """
    Download and save assets like images, CSS, JS files.
    """
    session=requests.Session()
    downloaded_assets = downloaded_assets or set()
    asset_url = normalize_url(asset_url, base_url)
    if asset_url in list(downloaded_assets):
        return
    downloaded_assets.add(asset_url)

    parsed_url = urlparse(asset_url)
    asset_path = parsed_url.path.lstrip('/')
    if not asset_path:
        return  # Skip if asset path is empty

    asset_full_path = os.path.join(output_dir, asset_path)
    return asset_full_path
def save_asset(asset_url,
               base_url,
               output_dir,
               downloaded_assets=None,
               session=None):
    asset_full_path = get_asset_path(asset_url=asset_url,
                                     base_url=base_url,
                                     output_dir=output_dir,
                                     downloaded_assets=downloaded_assets,
                                     session=session)
    if asset_full_path:
        os.makedirs(os.path.dirname(asset_full_path), exist_ok=True)

        try:
            response = session.get(asset_url, stream=True)
            response.raise_for_status()
            with open(asset_full_path, 'wb') as f:
                shutil.copyfileobj(response.raw, f)
            print(f"Saved asset: {asset_full_path}")
        except Exception as e:
            print(f"Failed to save asset {asset_url}: {e}")
        return downloaded_assets



# ssl_manager.py
from ..imports import *

class CipherManager:
    @staticmethod
    def get_default_ciphers() -> list:
        return ['ECDHE-ECDSA-AES128-SHA256',
                'ECDHE-RSA-AES256-SHA',
                'ECDHE-RSA-AES256-SHA384',
                'ECDHE-RSA-CHACHA20-POLY1305',
                'ECDHE-ECDSA-CHACHA20-POLY1305',
                'ECDHE-ECDSA-AES256-SHA384',
                'ECDHE-ECDSA-AES128-GCM-SHA256',
                'ECDHE-RSA-AES128-GCM-SHA256',
                'AES256-SHA',
                'AES128-SHA',
                'ECDHE-RSA-AES128-SHA256',
                'ECDHE-ECDSA-AES256-SHA',
                'ECDHE-RSA-AES256-GCM-SHA384',
                'ECDHE-ECDSA-AES256-GCM-SHA384']

    def __init__(self, cipher_list=None):
        self.cipher_list = cipher_list or self.get_default_ciphers()
        if isinstance(self.cipher_list, str):
            self.cipher_list = [c.strip() for c in self.cipher_list.split(',') if c.strip()]
        self.ciphers_string = ','.join(self.cipher_list) if self.cipher_list else ''
    def add_string_list(self):
        if len(self.cipher_list)==0:
            return ''
        return','.join(self.cipher_list)
    def create_list(self):
        if self.cipher_list == None:
            self.cipher_list= []
        elif isinstance(self.cipher_list, str):
            self.cipher_list=self.cipher_list.split(',')
        if isinstance(self.cipher_list, str):
            self.cipher_list=[self.cipher_list]
class CipherManagerSingleton:
    _instance = None
    @staticmethod
    def get_instance(cipher_list=None):
        if CipherManagerSingleton._instance is None:
            CipherManagerSingleton._instance = CipherManager(cipher_list=cipher_list)
        elif CipherManagerSingleton._instance.cipher_list != cipher_list:
            CipherManagerSingleton._instance = CipherManager(cipher_list=cipher_list)
        return CipherManagerSingleton._instance



from ..imports import *
from .functions import *
from .functions import _free_port,_make_chrome_options,_looks_like_html,_requests_fallback,_wait_until_ready,_make_profile_dir
from .seleneumManager import *
def _auto_age_gate(driver):
    try:
        yes = driver.find_element(By.ID, "age-button-yes")
        yes.click()
        time.sleep(0.3)
        return True
    except Exception:
        return False


# ---- Hardened page-source retrieval with fallback ----
def get_selenium_source(url, max_retries: int = 2, request_fallback: bool = True, timeout: float = 12.0):
    url_mgr = urlManager(url)
    if not url_mgr.url:
        return None
    url = str(url_mgr.url)

    manager = seleneumManager(url)
    key, driver = manager.get_driver(url)
    driver.get(url)
    _wait_until_ready(driver, timeout=timeout)
    last_exc = None
    try:
        for attempt in range(1, max_retries + 1):
            try:
                driver.get(url)
                _wait_until_ready(driver, timeout=timeout)
                html = driver.page_source or ""
                if not _looks_like_html(html):
                    html = driver.execute_script(
                        "return document.documentElement ? document.documentElement.outerHTML : '';"
                    ) or html
                if _looks_like_html(html):
                    return html
                logging.warning(f"Selenium returned suspicious HTML (len={len(html)}) for {url} "
                                f"[attempt {attempt}/{max_retries}]")
            except Exception as e:
                last_exc = e
                logging.warning(f"Selenium attempt {attempt}/{max_retries} failed for {url}: {e}")
            time.sleep(0.5 * attempt)

        if request_fallback:
            resp = _requests_fallback(url, headers={"User-Agent": "Mozilla/5.0"})
            if resp is not None:
                ctype = (resp.headers.get("content-type") or "").lower()
                body = resp.text if hasattr(resp, "text") else (
                    resp.content.decode("utf-8", "ignore") if hasattr(resp, "content") else ""
                )
                if "application/json" in ctype:
                    try:
                        return json.dumps(resp.json())
                    except Exception:
                        return body
                return body if _looks_like_html(body) or body else None
    finally:
        # critical: release the user-data-dir to avoid “already in use”
        manager.close_driver(key)

    if last_exc:
        logging.error(f"Unable to retrieve page for {url}: {last_exc}")
    return None

def get_driver(self, url):
    from ..imports.constants import _resolve_chrome_binary
    from selenium.webdriver.chrome.service import Service
    bin_path = _resolve_chrome_binary()
    opts, prof = _make_chrome_options(binary_path=bin_path, user_data_dir=None)
    service = Service('/usr/local/bin/chromedriver')
    driver = webdriver.Chrome(service=service, options=opts)
    key = f"{url}#{time.time()}"
    seleneumManager()._sessions[key] = {"driver": driver, "profile": prof}
    return key, driver



from .init_imports import *
from .module_imports import *
VIDEO_ENV_KEY = "DATA_DIRECTORY"
# --- hard defaults ---
VIDEOS_ROOT_DEFAULT = "~/videos"
DOCS_ROOT_DEFAULT   = "~/Documents"
DEFAULT_DOCUMENTS_ROOT = DOCS_ROOT_DEFAULT
DEFAULT_VIDEOS_ROOT = VIDEOS_ROOT_DEFAULT
load()  # load once
DEFINITIION_ALIAS = {
    "text":{"alias":["text","post","content","tweet","thread","body"],"description":"The main text."},
    "url":{"alias":["url","u","link","site","domain","intentUrl","link_address","address","canonical",'path'],"description":"A URL to include in the post."},
    "via":{"alias":["via","username","user","@","uploader","author"],"description":"A username to attribute the post to."},
    "hashtags":{"alias":["hashtags","hashtag","tag","tags"],"description":"Comma-separated hashtags (without the # symbol)."},
    "related":{"alias":["related","contributor","credit","cc","bcc"],"description":"Comma-separated related accounts."},
    "subject":{"alias":["title","subject","heading","header"],"description":"The email subject."},
    "body":{"alias":["text","post","content","tweet","thread","body"],"description":"The email body."},
    "cc":{"alias":["related","contributor","credit","cc","bcc"],"description":"Additional email addresses."},
    "bcc":{"alias":["related","contributor","credit","cc","bcc"],"description":"Additional email addresses."}
}
def get_alias_params(string):
    return DEFINITIION_ALIAS.get(string)
SOCIAL_SHARE_PARAMS={
    "x":{
        "url":"https://twitter.com/intent/tweet",
        "params":{
            "text":get_alias_params("text"),
            "url":get_alias_params("url"),
            "via":get_alias_params("via"),
            "hashtags":get_alias_params("hashtags"),
            "related":get_alias_params("related")
            },
        "characters":{"limit":280,"optimal":100,"mobile_cutoff":150,"url_len":30},
        "alias":["x","twitter","x.com","tweet","twitter.com"],
        "hash_symbol":False
        },
     "facebook":{
         "url":"http://facebook.com/sharer.php",
         "params":{
             "u":get_alias_params("url")
             },
         "characters":{"limit":63206,"optimal":50,"mobile_cutoff":150,"url_len":None},
         "alias":["facebook","fb","facebook.com","meta","meta.com"],
         "hash_symbol":True
         },
    
     "threads":{
         "url":"https://www.threads.net/intent/post",
         "params":{
             "text":get_alias_params("text")
             },
         "characters":{"limit":500,"optimal":150,"mobile_cutoff":150,"url_len":None},
         "alias":["threads","@","threads.com","@.com"],
         "hash_symbol":True
         },
    "mailto":{
        "url":"mailto:",
        "params":{
             "subject":get_alias_params("subject"),
             "body":get_alias_params("body"),
             "cc":get_alias_params("cc"),
             "bcc":get_alias_params("bcc")
             },
        "characters":{"limit":None,"optimal":None,"mobile_cutoff":None},
        "alias":["mailto","mail","email","email.com","mail.com"],
        "hash_symbol":True
        },
    "minds":{
        "url":"https://www.minds.com/newsfeed/subscriptions/latest",
        "params":{
            "intentUrl":get_alias_params("url")
            },
        "characters":{"limit":500,"optimal":125,"mobile_cutoff":150,"url_len":None},
        "alias":["minds","mindscollective","collective"],
        "hash_symbol":True
        }
             
    }

META_VARS = {
    "title": {"min": [20, 30], "max": [50,60]},
    "description": {"min": [70, 100], "max": [150, 160]},
    "alt": {"min": [10, 20], "max": [100, 125]},
    "context": {"min": [20, 30], "max": [70, 100]},
    "keywords": {"min": [10, 20], "max": [200, 250]}
}
ATTR_RE = re.compile(r'([a-zA-Z0-9:_-]+)\s*=\s*([\'"`])([^\'"`]+)\2')
TAG_OPEN_RE = re.compile(r'<\s*([a-zA-Z0-9:_-]+)')
TAG_CONTENT_RE = re.compile(r'<\s*([a-zA-Z0-9:_-]+)[^>]*>(.*?)</\s*\1\s*>', re.DOTALL)

EXTENTIONS=['.ac', '.academy', '.accountant', '.actor', '.agency', '.ai', '.airforce', '.am',
            '.apartments', '.archi', '.army', '.art', '.asia', '.associates', '.at', '.attorney',
            '.auction', '.audio', '.baby', '.band', '.bar', '.bargains', '.be', '.beer', '.berlin',
            '.best', '.bet', '.bid', '.bike', '.bingo', '.bio', '.biz', '.black', '.blackfriday',
            '.blog', '.blue', '.boston', '.boutique', '.br.com', '.build', '.builders', '.business',
            '.buzz', '.buz', '.ca', '.cab', '.cafe', '.camera', '.camp', '.capital', '.cards', '.care',
            '.careers', '.casa', '.cash', '.casino', '.catering', '.cc', '.center', '.ceo', '.ch', '.charity',
            '.chat', '.cheap', '.christmas', '.church', '.city', '.claims', '.cleaning', '.click', '.clinic',
            '.clothing', '.cloud', '.club', '.cn.com', '.co', '.co.com', '.co.in', '.co.nz', '.co.uk', '.coach',
            '.codes', '.coffee', '.college', '.com', '.com.co', '.com.mx', '.com.tw', '.community', '.company',
            '.computer', '.condos', '.construction', '.consulting', '.contact', '.contractors', '.cooking',
            '.cool', '.coupons', '.courses', '.credit', '.creditcard', '.cricket', '.cruises', '.cymru',
            '.cz', '.dance', '.date', '.dating', '.de', '.de.com', '.deals', '.degree', '.delivery',
            '.democrat', '.dental', '.dentist', '.desi', '.design', '.diamonds', '.diet', '.digital',
            '.direct', '.directory', '.discount', '.doctor', '.dog', '.domains', '.download', '.earth',
            '.eco', '.education', '.email', '.energy', '.engineer', '.engineering', '.enterprises',
            '.equipment', '.estate', '.eu', '.eu.com', '.events', '.exchange', '.expert', '.exposed',
            '.express', '.fail', '.faith', '.family', '.fans', '.farm', '.fashion', '.film', '.finance',
            '.financial', '.fish', '.fishing', '.fit', '.fitness', '.flights', '.florist', '.flowers', '.fm',
            '.football', '.forsale', '.foundation', '.fun', '.fund', '.furniture', '.futbol', '.fyi', '.gallery',
            '.games', '.garden', '.gay', '.gift', '.gifts', '.gives', '.glass', '.global', '.gmbh', '.gold',
            '.golf', '.graphics', '.gratis', '.green', '.gripe', '.group', '.gs', '.guide', '.guitars', '.guru',
            '.haus', '.healthcare', '.help', '.hiphop', '.hn', '.hockey', '.holdings', '.holiday', '.horse',
            '.host', '.hosting', '.house', '.how', '.immo', '.in', '.industries', '.info', '.ink', '.institue',
            '.insure', '.international', '.investments', '.io', '.irish', '.it', '.jetzt', '.jewelry', '.jp',
            '.jpn.com', '.juegos', '.kaufen', '.kim', '.kitchen', '.kiwi', '.la', '.land', '.lawyer', '.lease',
            '.legal', '.lgbt', '.li', '.life', '.lighting', '.limited', '.limo', '.link', '.live', '.llc', '.loan',
            '.loans', '.lol', '.london', '.love', '.ltd', '.luxury ', '.maison', '.managment', '.market', '.marketing',
            '.mba', '.me', '.me.uk', '.media', '.memorial', '.men', '.menu', '.miami', '.mobi', '.moda', '.moe', '.money',
            '.monster', '.mortgage', '.mx', '.nagoya', '.navy', '.net', '.net.co', '.network', '.news', '.ngo', '.ninja',
            '.nl', '.nyc', '.okinawa', '.one', '.ong', '.online', '.org', '.org.in', '.org.uk', '.partners', '.parts',
            '.party', '.pet', '.ph', '.photo', '.photography', '.photos', '.physio', '.pics', '.pictures', '.pink',
            '.pizza', '.pl', '.place', '.plumbing', '.plus', '.poker', '.press', '.pro', '.productions', '.promo',
            '.properties', '.property', '.pub', '.qpon', '.quebec', '.racing', '.realty', '.recipes', '.red', '.rehab',
            '.reisen', '.rent', '.rentals', '.repair', '.report', '.republican', '.rest', '.restaurant', '.review',
            '.reviews', '.rip', '.rocks', '.rodeo', '.run', '.sa.com', '.sale', '.sarl', '.sc', '.school', '.schule',
            '.science', '.se.net', '.services', '.sexy', '.sg', '.shiksha', '.shoes', '.shop', '.shopping', '.show',
            '.singles', '.site', '.ski', '.soccer', '.social', '.software', '.solar', '.solutions', '.soy', '.space',
            '.srl', '.store', '.stream', '.studio', '.study', '.style', '.supplies', '.supply', '.support', '.surf',
            '.surgery', '.systems', '.tattoo', '.tax', '.taxi', '.team', '.tech', '.technology', '.tel', '.tennis',
            '.theater', '.tienda', '.tips', '.today', '.tokyo', '.tools', '.tours', '.town', '.toys', '.trade', '.training',
            '.tv', '.tw', '.uk', '.uk.com', '.university', '.uno', '.us', '.us.com', '.vacations', '.vc', '.vegas',
            '.ventures', '.vet', '.viajes', '.video', '.villas', '.vip', '.vision', '.vodka', '.vote', '.voting',
            '.voyage', '.watch', '.webcam', '.website', '.wedding', '.wiki', '.win', '.wine', '.work', '.works',
            '.world', '.ws', '.wtf', '.xyz', '.yoga', '.za.com', '.zone']
POPULAR_EXTENTIONS = [
    '.com','.io','.ai','.net','.org','.co','.us'
    ]
ALL_EXTENTIONS = list(set(POPULAR_EXTENTIONS+EXTENTIONS))
ALL_URL_KEYS = {
    'scheme':['https','http'],
    'netloc':{
        "www":[True,False],
        "extentions":[POPULAR_EXTENTIONS,ALL_EXTENTIONS]
        }
    }
EXTENSIONS = EXTENTIONS
POPULAR_EXTENSIONS = POPULAR_EXTENTIONS
ALL_EXTENSIONS = ALL_EXTENTIONS
INVERSE_HTTP = {'http': 'https', 'https': 'http'}
INVERSE_BOOL = {True:False,False:True}

try:
    ENC = tiktoken.get_encoding("cl100k_base")
except Exception:
    print("⚠️ Falling back to local encoding.")
    ENC = tiktoken.Encoding(
        name="fallback",
        pat_str=r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+""",
        mergeable_ranks={},
        special_tokens={"": 100257}
    )

YOUTUBE_HOSTS = {"youtube.com", "www.youtube.com", "youtu.be"}
VIMEO_HOSTS = {"vimeo.com", "www.vimeo.com"}
TIKTOK_HOSTS = {"tiktok.com", "www.tiktok.com"}
TWITTER_HOSTS = {"twitter.com", "x.com", "www.twitter.com", "www.x.com"}
DIRECT_EXTS = {
    ".png", ".jpg", ".jpeg", ".gif", ".webp", ".svg",
    ".mp4", ".mov", ".mkv", ".avi",
    ".mp3", ".wav", ".ogg",
    ".pdf", ".zip", ".rar", ".7z", ".gz",
}
# -----------------------------------------------------
# DIRECT FILE DETECTION
# -----------------------------------------------------
FACEBOOK_HOSTS = {
    "facebook.com",
    "www.facebook.com",
    "fb.watch",
    "m.facebook.com"
}



from .imports import *
import json,re,urllib,webbrowser
from urllib.parse import urlparse, parse_qs, parse_qsl
from .seleneum import *
from abstract_utilities import get_home_dir
def determine_matching_dict(dict_og,dict_comp):
    for key,value in dict_og.items():
        if key not in dict_comp or dict_comp.get(key) != value:
            return False
    return True
def is_header_matching(headers_list,headers_comp):
    for headers in headers_list:
        if determine_matching_dict(headers,headers_comp):
            return True
    return False
def get_unique_headers_list(existing_headers, new_headers):
    result = list(existing_headers)  # copy
    for candidate in new_headers:
        if not is_header_matching(result, candidate):
            result.append(candidate)
    return result

def get_approved_headers_file_path(directory=None,file_path=None,basename=None):
    basename = basename or 'approved_headers.json'
    if file_path:
        return file_path
    if directory and os.path.isdir(directory):
        file_path = os.path.join(directory,basename)
    else:
        file_path = os.path.join(get_home_dir(),basename)
    return file_path
def save_approved_headers_dict(data,directory=None,file_path=None,basename=None):
    approved_headers_file_path = get_approved_headers_file_path(directory=directory,file_path=file_path,basename=basename)
    safe_dump_to_json(data=data,file_path=approved_headers_file_path)
def update_approved_headers(obj, url=None, directory=None, file_path=None, basename=None):
    approved_headers_dict = get_approved_headers_dict(
        directory=directory,
        file_path=file_path,
        basename=basename
    )
    if isinstance(obj, list):
        if not url:
            raise ValueError("url is required when obj is a list")
        domain = urlManager(url).parsed.get("full_domain")
        obj = {domain: obj}

    for domain, values in obj.items():
        current = approved_headers_dict.get(domain, [])
        approved_headers_dict[domain] = get_unique_headers_list(current, values)

    save_approved_headers_dict(
        approved_headers_dict,
        directory=directory,
        file_path=file_path,
        basename=basename
    )
def get_approved_headers_dict(directory=None,file_path=None,basename=None):
    approved_headers_file_path = get_approved_headers_file_path(directory=directory,file_path=file_path,basename=basename)
    if not os.path.isfile(approved_headers_file_path):
        safe_dump_to_json(data={},file_path=approved_headers_file_path)
    return safe_load_from_json(approved_headers_file_path)
def get_approved_headers_for_domain(url):
    domain = urlManager(url).parsed.get("domain")
    approved_headers_dict = get_approved_headers_dict()
    return approved_headers_dict.get(domain, [])
def get_approved_header_for_domain(url,restricted_list=None):
    restricted_list = restricted_list or []
    approved_headers = get_approved_headers_for_domain(url)
    for approved_header in approved_headers:
        if not is_header_matching(approved_header,restricted_list):
            return approved_header
    return derive_approved_headers_for_url(url,return_first_found=True,restricted_list=restricted_list)
def get_check_approved_header(url):
    if not is_header_matching(user_agent_list,headers) and is_header_valid_for_url(url,user_agent,headers):
        user_agent_list.append(headers)
        update_approved_headers(user_agent_list,url=url)
def is_valid_request_response(source_code):
    text = str(source_code)
    txts = [
        "not a bot",
        'update your browser',
        'Bitte aktualisiere deinen BrowserDein Browser',
        'browser is no longer supported',
        "Dein Browser"
        ]
    for txt in txts:
        if txt not in text:
            return False
    return True
def is_header_valid_for_url(url,user_agent,headers,session):
    source_code = fetch_html(
                url=url,
                session=session,
                headers=headers,
                user_agent=user_agent
            )
    if is_valid_request_response(source_code):
        return source_code
def generate_ua_and_headers(url):
    ua_mgr = UserAgentManager(randomAll=True)
    user_agent = ua_mgr.get_user_agent()
    headers = ua_mgr.generate_headers()
    return user_agent,headers 
def generate_ua_and_headers_session(url):
    user_agent,headers = generate_ua_and_headers(url)
    session = requests.Session()
    return user_agent,headers ,session
def derive_approved_headers_for_url(url,return_first_found=True,restricted_list=None):
    restricted_list=restricted_list or []
    user_agent_list = get_approved_headers_for_domain(url)
    while True:
        user_agent,headers,session = generate_ua_and_headers_session(url)
        header_matching = is_header_matching(user_agent_list+restricted_list,headers)
        if not header_matching:
            source_code = is_header_valid_for_url(url,user_agent,headers,session)
            if source_code:
                user_agent_list.append(headers)
                update_approved_headers(user_agent_list,url=url)
                if return_first_found:
                    return headers
def derive_approved_headers_user_agent_session_for_url(url,return_first_found=True,restricted_list=None):
    restricted_list=restricted_list or []
    user_agent_list = get_approved_headers_for_domain(url)
    while True:
        user_agent,headers,session = generate_ua_and_headers_session(url)
        header_matching = is_header_matching(user_agent_list+restricted_list,headers)
        if not header_matching:
            source_code = is_header_valid_for_url(url,user_agent,headers,session)
            if source_code:
                user_agent_list.append(headers)
                update_approved_headers(user_agent_list,url=url)
                if return_first_found:
                    return user_agent,headers,session,source_code






class YouTubeResolveError(Exception):
    pass




# ---------------------------------------------------------
# MAIN
# ---------------------------------------------------------
if __name__ == "__main__":
    VIDEO_URL = "https://www.youtube.com/shorts/6vP02wYh4Ds"
    player = extract_player_response(VIDEO_URL)
    for itag in get_any_value(player,'itag'):
    
        downloader = AbstractYouTubeDownloader()

        print("[*] Resolving direct URL...")
        result = downloader.resolve_direc



from .abstractVideoManager import AbstractVideoManager
from .extract import (
    extract_player_response,
    extract_googlevideo_urls_from_html,
    iter_streaming_urls,
)
from .validators import looks_like_real_stream
from .seleneum_utils import *



from .imports import *
from .functions import *
from urllib3.util.retry import Retry               # <-- correct place
import logging
class NetworkManager:
    def __init__(self, user_agent_manager=None, user_agent=None, tls_adapter=None, ssl_manager=None,
                 proxies=None, cookies=None, ciphers=None, certification: Optional[str]=None,
                 ssl_options: Optional[List[str]]=None):
        self.ua_mgr = user_agent_manager or UserAgentManager()
        self.ciphers = ciphers or CipherManager().ciphers_string
        self.ssl_mgr = ssl_manager or SSLManager(
            ciphers=self.ciphers or CipherManager().ciphers_string,
            ssl_options=ssl_options,
            certification=certification
        )

        self.session = requests.Session()
        self.session.headers.update({
            "User-Agent": user_agent or self.ua_mgr.user_agent,
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.9",
            "Connection": "keep-alive"
        })

        self.tls_adapter = tls_adapter or TLSAdapter(self.ssl_mgr)
        self.session.mount("https://", self.tls_adapter)
        self.session.mount("http://", HTTPAdapter())

        if proxies:
            self.session.proxies = proxies
        self.proxies = self.session.proxies

        # normalize cookies then load into jar
        norm = self._normalize_cookies(cookies)
        if norm:
            jar = requests.cookies.RequestsCookieJar()
            for k, v in norm.items():
                jar.set(k, v)
            self.session.cookies = jar
        self.cookies = norm

        # robust retries
        retry = Retry(total=5, backoff_factor=0.5, status_forcelist=[429,500,502,503,504])
        self.session.get_adapter("https://").max_retries = retry
        self.session.get_adapter("http://").max_retries = retry

    def _normalize_cookies(self, cookies):
        """Ensure cookies are always a dict of str->str"""
        if not cookies:
            return {}
        if isinstance(cookies, dict):
            return {str(k): str(v) for k, v in cookies.items()}
        if isinstance(cookies, requests.cookies.RequestsCookieJar):
            return {c.name: str(c.value) for c in cookies}
        logging.warning("Dropping invalid cookies object: %r", cookies)
        return {}



# abstract_youtube/validators.py
from .imports import *
def looks_like_real_stream(url: str) -> bool:
    q = parse_qs(urlparse(url).query)
    return (
        "itag" in q and
        "mime" in q and
        ("sig" in q or "signature" in q)
    )



from ..imports import *
import os


def get_http(domain: str | None) -> str | None:
    """Return 'http' or 'https' if detected, otherwise None."""
    if not domain:
        return None
    domain = str(domain).strip()
    if domain.startswith("https"):
        return "https"
    if domain.startswith("http"):
        return "http"
    return None


def strip_http(domain: str | None, http: str | None = None) -> str | None:
    """Remove http(s):// from domain if present."""
    if not domain:
        return None
    http = http or get_http(domain)
    try:
        if http:
            domain = domain[len(http) + 3 :] if "://" in domain else domain[len(http):]
        domain = eatAll(domain, [":", "/"])
        return domain.strip()
    except Exception:
        return domain


def get_www(domain: str | None, http: str | None = None) -> bool:
    """Return True if domain starts with www."""
    if not domain:
        return False
    domain = strip_http(domain,http)
    return bool(domain and domain.startswith("www."))

def strip_www(domain: str | None, http: str | None = None) -> str | None:
    """Remove leading www. and restore http prefix if any."""
    if not domain:
        return None
    http = http or get_http(domain)
    domain = strip_http(domain, http)
    try:
        if domain and domain.startswith("www."):
            domain = domain[len("www.") :]
        if http:
            domain = f"{http}://{domain}"
        return domain.strip()
    except Exception:
        return domain



def get_http_www(domain: str | None) -> dict:
    """Return dict with detected http and www info."""
    if not domain:
        return {"http": None, "www": False}
    http = get_http(domain)
    www = get_www(domain, http)
    return {"http": http, "www": www}


def strip_http_www(domain) -> str | None:
    """Remove both http(s) and www."""
    if not domain:
        return None
    http = get_http(domain)
    domain = strip_www(domain, http)
    domain = strip_http(domain, http)
    return domain


def get_stripped_domain(domain: str | None, http: str | None = None) -> list[str] | None:
    """Return list of domain parts, safely."""
    if not domain:
        return None
    domain = strip_http_www(domain, http)
    domain,params = get_url_params(domain,clean_url=True)
    return domain.split("/") if domain else None

def get_domain_paths(domain: str | None, http: str | None = None) -> dict:
    """Return {'domain': <domain>, 'path': <path>} even if None."""
    if not domain:
        return {"domain": None, "path": ""}

    return get_url_paths_params(domain, http).get("paths")





def get_domain_path(domain: str | None, http: str | None = None) -> str:
    """Extract path component from URL."""
    domain_paths = get_domain_paths(domain, http=http)
    return '/'.join(domain_paths)


def get_domain_name_ext(domain: str | None, http: str | None = None) -> dict:
    """Return {'name': name, 'ext': ext} safely."""
    if not domain:
        return {"name": None, "ext": None}
    domain_only = get_domain(domain, http)
    if not domain_only:
        return {"name": None, "ext": None}
    name, ext = os.path.splitext(domain_only)
    return {"name": name or None, "ext": ext or None}




def get_url_paths_params(domain,http=None):

    domain_vars = {}
    domain_vars["original_url"] = domain
    domain_vars["http"]= get_http(domain=domain)
    domain_vars["www"]= get_www(domain=domain,http=domain_vars.get("http"))
    
    
    domain_spl = domain.split('?')
    domain_vars['clean_url']= domain_spl[0]
    params= {}
    domain_vars['param']=''
    if len(domain_spl)>1:
        url = domain_spl[0]
        raw_params = domain_spl[-1]
        domain_vars['param'] = raw_params
        for raw_param in raw_params.split('&'):
            key_value = raw_param.split('=')
            key = key_value[0]
            domain_vars[key]= None
            if len(key_value)>1:
                domain_vars[key]= '='.join(key_value[1:])
    domain_vars['params']=params
    
    stripped_url = strip_http_www(domain=domain)

    domain_spl = stripped_url.split("/")
    domain = domain_spl[0]
    domain_vars['domain'] = domain
    name,ext = os.path.splitext(domain)
    domain_vars['name'] = name
    domain_vars['ext'] = ext
    paths = []
    if len(domain_spl)>1:
        paths = domain_spl[1:]
    domain_vars['paths'] = [path for path in paths if path]
    domain_vars['path'] = '/'.join(domain_vars['paths'])
    return domain_vars
def get_extension(domain: str | None = None, http: str | None = None, parsed=None, options=None) -> str | None:
    """Get file extension safely, even if parsed URL object is missing."""
    try:
        if parsed:
            domain = getattr(parsed, "netloc", domain)
            http = getattr(parsed, "scheme", http)
    except Exception:
        pass
    return get_url_paths_params(domain=domain, http=http).get("ext")


def get_domain_name(domain: str | None, http: str | None = None) -> str | None:
    """Get domain name without extension."""
    return get_url_paths_params(domain=domain, http=http).get("name")
def get_url_path(domain: str | None, http: str | None = None):
    return get_url_paths_params(domain=domain, http=http).get("path")
def get_url_paths(domain: str | None, http: str | None = None):
    return get_url_paths_params(domain=domain, http=http).get("paths")
def get_cleaned_url(domain: str | None, http: str | None = None):
    return get_url_paths_params(domain=domain, http=http).get("clean_url")
def get_domain(domain: str | None, http: str | None = None) -> str | None:
    """Extract bare domain from URL."""
    return get_url_paths_params(domain, http).get("domain")
def get_params(domain: str | None, http: str | None = None) -> str | None:
    """Extract bare domain from URL."""
    return get_url_paths_params(domain, http).get("params")
def get_param(domain: str | None, http: str | None = None) -> str | None:
    """Extract bare domain from URL."""
    return get_url_paths_params(domain, http).get("param")
def get_video_url(url=None,video_url=None):
    return get_cleaned_url(video_url or url)
    
# Backward compatibility alias

get_extention = get_extension




from ..src import *
def _tiktoken_split(word: str):
    token_ids = ENC.encode(word)
    pieces = [ENC.decode([tid]) for tid in token_ids]
    merged, buf = [], ""

    for p in pieces:
        if p.isalpha():
            buf += p
        else:
            if buf:
                merged.append(buf)
                buf = ""
    if buf:
        merged.append(buf)

    return merged or [word]

def tokenize_string(domain: str):
    primary=[""]
    if domain:
        domain = domain.lower().strip()
        root = re.sub(r"\.[a-z0-9]{2,6}$", "", domain)

        primary = segment(root)

        if len(primary) <= 1:
            return _tiktoken_split(root)

    return primary
tokenize_domain = tokenize_string



from .DomainManager import *
from .url_utils import *
from ..imports import *


def get_parsed_dict(url=None, parsed=None, parsed_dict=None):
    if url and not parsed:
        parsed = parse_url(url)
    if parsed and not parsed_dict:
        parsed_dict = parse_url(parsed=parsed)
    return parsed_dict 
def reconstructUrlFromUrlParse(url=None, parsed=None, parsed_dict=None):
    parsed_dict = parsed_dict or get_parsed_dict(url)
    keys = ['scheme','domain','netloc','path','params','query','fragment']
    if parsed_dict:
        scheme = parsed_dict.get('scheme')
        
        nuUrl = ''
        for key in keys:
            value = parsed_dict.get(key, '')
            if key == 'scheme':
                nuUrl += f'{value}://' if value else ''
            elif key == 'query':
                nuUrl += f'?{reconstructQuery(value)}' if value else ''
            elif key == 'netloc':
                nuUrl += reconstructNetLoc(value) if value else ''
            else:
                # removed noisy print(value)
                nuUrl += f"/{value}"
        return nuUrl
    return url
def get_youtube_parsed_dict(url=None, parsed=None, parsed_dict=None):
    parsed_dict =  get_parsed_dict(url=url, parsed=parsed, parsed_dict=parsed_dict)
    if parsed_dict:
        netloc = parsed_dict.get("netloc")
        domain = (netloc or {})
        if isinstance(domain,dict):
            domain = domain.get('domain') or ''
        query  = parsed_dict.get('query') or {}
        path   = parsed_dict.get('path') or ''
        if domain.startswith('youtu'):
            # force youtube.com and /watch?v=ID
            netloc['www'] = True
            netloc['domain'] = 'youtube'
            netloc['extention'] = '.com'
            parsed_dict['netloc'] = netloc

            # keep v if present; otherwise derive
            v_query = query.get('v')
            if not v_query:
                if path.startswith('/watch/') or path.startswith('/shorts/'):
                    v_query = eatAll(path, ['/','watch','shorts'])
                else:
                    v_query = eatAll(path, ['/'])
            parsed_dict['path'] = '/watch'
            parsed_dict['query'] = {'v': v_query} if v_query else {}
            return parsed_dict
def get_youtube_v_query(url=None, parsed=None, parsed_dict=None):
    parsed_dict =  get_youtube_parsed_dict(url=url, parsed=parsed, parsed_dict=parsed_dict)
    v_query = parsed_dict.get('query',{})
    if isinstance(v_query,dict):
        v_query = v_query.get('v')
    return v_query
def get_youtube_url(url=None, parsed=None, parsed_dict=None):
    parsed_dict =  get_youtube_parsed_dict(url=url, parsed=parsed, parsed_dict=parsed_dict)
    return reconstructUrlFromUrlParse(parsed_dict=parsed_dict)
def get_threads_url(url=None,parsed=None, parsed_dict =None ):
    parsed_dict =  get_parsed_dict(url=url, parsed=parsed, parsed_dict=parsed_dict)
    if parsed_dict:
        netloc = parsed_dict.get("netloc")
        domain = (netloc or {})
        if isinstance(domain,dict):
            domain = domain.get('domain') or ''
        if domain.startswith('threads'):
            netloc['www']=True
            netloc['domain'] ='threads'
            netloc['extention'] = '.net'
            parsed['netloc']=netloc
            return reconstructUrlFromUrlParse(url=url,parsed=parsed,parsed_dict=None)  


def _clean_url(url: str) -> str:
    """ normalize scheme + strip weird fragments """
    if not url:
        return ""
    url = url.strip()
    if url.startswith("//"):
        url = "https:" + url
    return url

# -----------------------------------------------------
# PLATFORM EXTRACTORS
# -----------------------------------------------------

def detect_youtube(parsed, query):
    netloc = parsed.netloc.lower()
    if netloc not in YOUTUBE_HOSTS:
        return None

    # video id
    if "v" in query:
        return {
            "downloadable": True,
            "kind": "video",
            "provider": "youtube",
            "id": query.get("v", [None])[0],
            "direct": False,
        }

    # youtu.be short links
    if parsed.netloc == "youtu.be":
        vid = parsed.path.strip("/")
        return {
            "downloadable": True,
            "kind": "video",
            "provider": "youtube",
            "id": vid,
            "direct": False,
        }

    # playlists
    if "list" in query:
        return {
            "downloadable": True,
            "kind": "playlist",
            "provider": "youtube",
            "id": query.get("list", [None])[0],
            "direct": False,
        }

    return {
        "downloadable": False,
        "kind": None,
        "provider": "youtube",
        "id": None,
        "direct": False,
    }


def detect_vimeo(parsed):
    if parsed.netloc.lower() not in VIMEO_HOSTS:
        return None
    vid = parsed.path.strip("/")
    if vid.isdigit():
        return {
            "downloadable": True,
            "kind": "video",
            "provider": "vimeo",
            "id": vid,
            "direct": False
        }
    return {"downloadable": False, "provider": "vimeo"}


def detect_tiktok(parsed):
    if parsed.netloc.lower() not in TIKTOK_HOSTS:
        return None

    parts = parsed.path.split("/")
    # /@user/video/123456
    if "video" in parts:
        idx = parts.index("video")
        vid = parts[idx + 1] if idx + 1 < len(parts) else None
        return {
            "downloadable": True,
            "kind": "video",
            "provider": "tiktok",
            "id": vid,
            "direct": False,
        }
    return {"downloadable": False, "provider": "tiktok"}


def detect_twitter(parsed, query):
    if parsed.netloc.lower() not in TWITTER_HOSTS:
        return None

    parts = parsed.path.split("/")
    # /username/status/ID
    if "status" in parts:
        idx = parts.index("status")
        tid = parts[idx + 1] if idx + 1 < len(parts) else None
        return {
            "downloadable": True,
            "kind": "video",
            "provider": "twitter",
            "id": tid,
            "direct": False,
        }
    return {"downloadable": False, "provider": "twitter"}


def detect_facebook(parsed):
    net = parsed.netloc.lower()

    if net not in FACEBOOK_HOSTS:
        return None

    # Pattern 1: /share/v/{video_id}/
    # Example: https://www.facebook.com/share/v/1EqXAsf57B/
    parts = [p for p in parsed.path.split("/") if p]

    if len(parts) >= 3 and parts[0] == "share" and parts[1] == "v":
        vid = parts[2]
        return {
            "downloadable": True,
            "kind": "video",
            "provider": "facebook",
            "id": vid,
            "direct": False,
        }

    # Pattern 2: Watch URLs (fb.watch/xxxx)
    if net == "fb.watch":
        vid = parsed.path.strip("/")
        return {
            "downloadable": True,
            "kind": "video",
            "provider": "facebook",
            "id": vid,
            "direct": False,
        }

    # Pattern 3: Regular watch URLs
    # https://www.facebook.com/watch/?v=123456789
    query = parse_qs(parsed.query)
    if "v" in query:
        return {
            "downloadable": True,
            "kind": "video",
            "provider": "facebook",
            "id": query["v"][0],
            "direct": False,
        }

    return {"downloadable": False, "provider": "facebook"}
def detect_direct_file(parsed):
    ext = os.path.splitext(parsed.path)[1].lower()
    if ext in DIRECT_EXTS:
        return {
            "downloadable": True,
            "kind": mimetypes.guess_type(parsed.path)[0] or "file",
            "provider": None,
            "id": None,
            "direct": True,
        }
    return None


# -----------------------------------------------------
# MAIN ENTRY POINT
# -----------------------------------------------------

def get_downloadable_info(url: str):
    url = _clean_url(url)
    parsed = urlparse(url)
    query = parse_qs(parsed.query)

    # 1. Direct file
    direct = detect_direct_file(parsed)
    if direct:
        direct["url"] = url
        return direct

    # 2. YouTube
    yt = detect_youtube(parsed, query)
    if yt:
        yt["url"] = url
        return yt

    # 3. Vimeo
    vimeo = detect_vimeo(parsed)
    if vimeo:
        vimeo["url"] = url
        return vimeo

    # 4. TikTok
    tiktok = detect_tiktok(parsed)
    if tiktok:
        tiktok["url"] = url
        return tiktok

    # 5. Twitter/X
    tw = detect_twitter(parsed, query)
    if tw:
        tw["url"] = url
        return tw

    # fallback
    return {
        "downloadable": False,
        "kind": None,
        "provider": None,
        "id": None,
        "direct": False,
        "url": url,
    }

def get_corrected_url(url=None,parsed=None, parsed_dict =None ):
    return get_downloadable_info(url).get('url')
##    parsed_dict =  get_parsed_dict(url=url, parsed=parsed, parsed_dict=parsed_dict)
##    if parsed_dict:
##        funcs = [get_threads_url,get_youtube_url,reconstructUrlFromUrlParse]
##        for func in funcs:
##            corrected_url = func(url=url,parsed=parsed,parsed_dict=parsed_dict)
##            if corrected_url:
##                return corrected_url




from bs4 import BeautifulSoup
from typing import Any, Dict, List, Optional
import requests
from .get_ua import *



def get_source_code(url: str, session: Optional[requests.Session] = None) -> str:
    session = session or get_session(url)
    return fetch_html(url=url, session=session)


def get_soup(url: str, session: Optional[requests.Session] = None) -> BeautifulSoup:
    source_code = get_source_code(url, session=session)
    return BeautifulSoup(source_code, "html.parser")


def get_soup_text(url: str, session: Optional[requests.Session] = None) -> str:
    source_soup = get_soup(url, session=session)
    return source_soup.get_text(separator="\n", strip=True)


def get_meta_tags_from_soup(soup: BeautifulSoup) -> List[Dict[str, Any]]:
    """
    Return every <meta> tag as a dictionary.

    Example output:
    [
        {"name": "description", "content": "..."},
        {"property": "og:title", "content": "..."},
        {"charset": "utf-8"}
    ]
    """
    return [dict(tag.attrs) for tag in soup.find_all("meta")]

def get_meta_tags_from_url(url,session: Optional[requests.Session] = None) -> Dict[str, str]:
    soup = get_soup(url=url, session=session)
    return get_meta_tags_from_soup(soup)
def get_meta_map_from_soup(soup: BeautifulSoup) -> Dict[str, str]:
    """
    Return useful meta tags keyed by name/property/http-equiv/charset.

    Example:
    {
        "description": "...",
        "og:title": "...",
        "twitter:card": "...",
        "charset": "utf-8"
    }
    """
    meta_map = {}

    for tag in soup.find_all("meta"):
        attrs = dict(tag.attrs)

        key = (
            attrs.get("name")
            or attrs.get("property")
            or attrs.get("http-equiv")
            or attrs.get("charset")
        )

        if not key:
            continue

        if "content" in attrs:
            meta_map[str(key)] = attrs["content"]
        else:
            meta_map[str(key)] = ""

    return meta_map

def get_meta_map_from_url(url,session: Optional[requests.Session] = None) -> Dict[str, str]:
    soup = get_soup(url=url, session=session)
    return get_meta_map_from_soup(soup)
def get_body_from_soup(soup: BeautifulSoup, as_text: bool = False) -> str:
    body = soup.body

    if body is None:
        return ""

    if as_text:
        return body.get_text(separator="\n", strip=True)

    return str(body)
def get_body_from_url(url,session: Optional[requests.Session] = None) -> Dict[str, str]:
    soup = get_soup(url=url, session=session)
    return get_body_from_soup(soup)

def get_title_from_soup(soup: BeautifulSoup) -> str:
    if soup.title and soup.title.string:
        return soup.title.string.strip()
    return ""
def get_title_from_url(url,session: Optional[requests.Session] = None) -> Dict[str, str]:
    soup = get_soup(url=url, session=session)
    return get_title_from_soup(soup)

def get_desired_content_from_soup(
    soup: BeautifulSoup,
    selectors: Optional[List[str]] = None,
    remove_selectors: Optional[List[str]] = None,
    as_text: bool = True,
) -> Dict[str, Any]:
    """
    Pull useful page content.

    selectors lets you target desired areas:
        ["article", "main", ".content", "#post"]

    remove_selectors strips junk:
        ["script", "style", "nav", "footer", "header", "aside"]
    """

    remove_selectors = remove_selectors or [
        "script",
        "style",
        "noscript",
        "svg",
        "nav",
        "footer",
        "header",
        "aside",
        "form",
    ]

    for selector in remove_selectors:
        for tag in soup.select(selector):
            tag.decompose()

    selected_blocks = []

    if selectors:
        for selector in selectors:
            selected_blocks.extend(soup.select(selector))
    else:
        selected_blocks = (
            soup.select("article")
            or soup.select("main")
            or soup.select("[role='main']")
            or [soup.body]
        )

    content_parts = []

    for block in selected_blocks:
        if not block:
            continue

        if as_text:
            content_parts.append(block.get_text(separator="\n", strip=True))
        else:
            content_parts.append(str(block))

    return {
        "title": get_title_from_soup(soup),
        "meta": get_meta_map_from_soup(soup),
        "meta_tags": get_meta_tags_from_soup(soup),
        "content": "\n\n".join(part for part in content_parts if part),
    }

def get_desired_content_from_url(url,session: Optional[requests.Session] = None) -> Dict[str, str]:
    soup = get_soup(url=url, session=session)
    return get_desired_content_from_soup(soup)
def extract_page_content(
    url: str,
    session: Optional[requests.Session] = None,
    selectors: Optional[List[str]] = None,
    as_text: bool = True,
) -> Dict[str, Any]:
    soup = get_soup(url, session=session)

    return {
        "url": url,
        "title": get_title_from_soup(soup),
        "meta": get_meta_map_from_soup(soup),
        "meta_tags": get_meta_tags_from_soup(soup),
        "body": get_body_from_soup(soup, as_text=as_text),
        "desired": get_desired_content_from_soup(
            soup,
            selectors=selectors,
            as_text=as_text,
        ),
    }



YOUTUBE_HOSTS = {"youtube.com", "www.youtube.com", "youtu.be"}

def detect_youtube(parsed, query):
    net = parsed.netloc.lower()
    if net not in YOUTUBE_HOSTS:
        return None

    if "v" in query:
        return {"downloadable": True, "kind": "video", "provider": "youtube",
                "id": query["v"][0], "direct": False}

    if net == "youtu.be":
        vid = parsed.path.strip("/")
        return {"downloadable": True, "kind": "video", "provider": "youtube",
                "id": vid, "direct": False}

    if "list" in query:
        return {"downloadable": True, "kind": "playlist", "provider": "youtube",
                "id": query["list"][0], "direct": False}

    return {"downloadable": False, "provider": "youtube"}



from .imports import *
from .src import *

# ---------------------------------------------------------
# MAIN
# ---------------------------------------------------------

def makeAbstractVideoManager(output_dir='downloads'):
    return AbstractVideoManager(output_dir)

def get_itags(video_url):
    player = extract_player_response(video_url)
    return get_any_value(player, 'itag')

def get_itag(video_url, itag=None):
    itags = get_itags(video_url)
    if itag and itag in itags:
        return itag
    return itags[0]

def getDirectUrlDict(video_url, itag=None, output_dir='downloads', manager=None):
    if manager is None:
        manager = makeAbstractVideoManager(output_dir)
    itag = get_itag(video_url, itag=itag)
    return manager.resolve_direct_url(video_url=video_url, itag=itag)

def getDirectUrl(video_url, itag=None, manager=None):
    result = getDirectUrlDict(video_url=video_url, itag=itag, manager=manager)
    return result["direct_url"]

def getMetaData(video_url, itag=None, manager=None):
    result = getDirectUrlDict(video_url=video_url, itag=itag, manager=manager)
    return result["metadata"]

def getTitle(video_url, itag=None, manager=None):
    result = getMetaData(video_url=video_url, itag=itag, manager=manager)
    return result.get("title", "video").replace("/", "_")

def getVideoFilename(video_url, itag=None, manager=None):
    return getTitle(video_url=video_url, itag=itag, manager=manager)
    
def getVideoBasename(video_url, itag=None,filename=None,ext=None, manager=None):
    filename = filename or getVideoFilename(video_url, itag=itag, manager=manager)
    ext= ext or '.mp4'
    ext = eatAll(ext,'.')
    return f"{filename}.{ext}"
def check_dir_for_basename(basename,directory):
    dirlist = os.listdir(directory)
    for i,item in enumerate(dirlist):
        if item == basename:
            return True
def getFilePath(video_url,  itag=None, output_dir=None,filename=None,ext=None, manager=None):
    basename = getVideoBasename(video_url, itag=itag,filename=filename,ext=ext, manager=manager)
    output_dir = output_dir or get_default_videos_dir()
    file_path = os.path.join(output_dir,basename)
    if os.path.isfile(file_path):
        dirlist = os.listdir(output_dir)
        file_part = get_file_parts(file_path)
        dirname = file_part.get('dirname')
        filename = file_part.get('filename')
        ext = file_part.get('ext')
        basename = file_part.get('basename')
        for i,item in enumerate(dirlist):
            if not check_dir_for_basename(basename,dirname):
                file_path = os.path.join(dirname,basename)
                break
            basename = f"{filename}_{i+1}{ext}"
    return file_path
def abstractVideoDownload(video_url, itag=None, output_dir=None,filename=None,ext=None, manager=None):
    output_dir = output_dir or get_default_videos_dir()
    
    if manager is None:
        manager = makeAbstractVideoManager(output_dir)
    getVideoFilename(video_url=video_url, itag=itag, manager=manager)
    direct_url = getDirectUrl(video_url=video_url, itag=itag, manager=manager)
    file_path = getFilePath(video_url=video_url, itag=itag,output_dir=output_dir, filename=filename,ext=ext,manager=manager)
    return manager.download(url=direct_url, filename=file_path)



# pip install selenium webdriver-manager beautifulsoup4
import json, re, time
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from webdriver_manager.chrome import ChromeDriverManager
from bs4 import BeautifulSoup

def make_driver():
    chrome_options = Options()
    chrome_options.add_argument("--headless=new")
    chrome_options.add_argument("--no-sandbox")
    chrome_options.add_argument("--disable-dev-shm-usage")
    chrome_options.add_argument("--disable-gpu")
    chrome_options.add_argument("--disable-extensions")
    chrome_options.add_argument("--window-size=1400,900")
    chrome_options.add_argument("--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                                "AppleWebKit/537.36 (KHTML, like Gecko) "
                                "Chrome/123.0.0.0 Safari/537.36")
    caps = {
        "goog:loggingPrefs": {"performance": "ALL"},
    }
    driver = webdriver.Chrome(ChromeDriverManager().install(),
                              options=chrome_options,
                              desired_capabilities=caps)
    return driver

def get_all_meta_selenium(url: str) -> dict:
    driver = make_driver()
    try:
        driver.get(url)
        time.sleep(2)  # let SSR/consent bits settle
        html = driver.page_source
        soup = BeautifulSoup(html, "html.parser")

        out = {"title": soup.title.text.strip() if soup.title else None,
               "meta": [], "links": [], "json_ld": [], "citation": {}}

        for m in soup.find_all("meta"):
            entry = {k: m.get(k) for k in ("name","property","http-equiv","itemprop","charset","content")}
            entry = {k: v for k, v in entry.items() if v}
            if entry:
                out["meta"].append(entry)
                k = (entry.get("name") or entry.get("property") or "").lower()
                if k.startswith("citation_") and "content" in entry:
                    out["citation"].setdefault(k, []).append(entry["content"])

        for l in soup.find_all("link"):
            rel = l.get("rel")
            if isinstance(rel, list): rel = " ".join(rel)
            out["links"].append({
                "rel": rel, "href": l.get("href"),
                "type": l.get("type"), "title": l.get("title"), "hreflang": l.get("hreflang")
            })

        for s in soup.find_all("script", type=re.compile(r"application/ld\+json", re.I)):
            txt = s.get_text(strip=True)
            try: out["json_ld"].append(json.loads(txt))
            except Exception: out["json_ld"].append({"raw": txt})

        # Example: collect request headers via CDP logs (optional)
        logs = driver.get_log("performance")
        # parse logs if you want request/response headers

        return out
    finally:
        driver.quit()




from abstract_webtools import *
from abstract_utilities import *
import re
import json
import requests
import urllib.parse
from typing import Tuple, List, Dict, Any
from pathlib import Path

approved_headers = [{
            "User-Agent": "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Maxthon/4.4.6.1000 Chrome/30.0.1599.101 Safari/537.36",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8",
            "Accept-Language": "de-DE,de;q=0.9,en;q=0.8",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-User": "?1",
            "Sec-Fetch-Dest": "document",
            "Sec-CH-UA": "\"Chromium\";v=\"116\", \"Brave\";v=\"116\", \"Not A;Brand\";v=\"99\"",
            "Sec-CH-UA-Platform": "macOS",
            "Sec-CH-UA-Mobile": "?0"
        },
        {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:39.0) Gecko/20100101 Firefox/39.0",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.8,es;q=0.6",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-User": "?1",
            "Sec-Fetch-Dest": "document",
            "Sec-CH-UA": "\"Chromium\";v=\"116\", \"Brave\";v=\"116\", \"Not A;Brand\";v=\"99\"",
            "Sec-CH-UA-Platform": "macOS",
            "Sec-CH-UA-Mobile": "?0"
        },
        {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2503.0 Safari/537.36",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8",
            "Accept-Language": "de-DE,de;q=0.9,en;q=0.8",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-User": "?1",
            "Sec-Fetch-Dest": "document",
            "Sec-CH-UA": "\"Chromium\";v=\"116\", \"Brave\";v=\"116\", \"Not A;Brand\";v=\"99\"",
            "Sec-CH-UA-Platform": "Windows",
            "Sec-CH-UA-Mobile": "?0"
        },
        {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:35.0) Gecko/20100101 Firefox/35.0",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.8,es;q=0.6",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-User": "?1",
            "Sec-Fetch-Dest": "document",
            "Sec-CH-UA": "\"Chromium\";v=\"118\", \"Google Chrome\";v=\"118\", \"Not A;Brand\";v=\"99\"",
            "Sec-CH-UA-Platform": "macOS",
            "Sec-CH-UA-Mobile": "?0"
        },
        {
            "User-Agent": "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2062.120 Safari/537.36",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.8,es;q=0.6",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "close",
            "Upgrade-Insecure-Requests": "1",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-User": "?1",
            "Sec-Fetch-Dest": "document",
            "Sec-CH-UA": "\"Chromium\";v=\"116\", \"Brave\";v=\"116\", \"Not A;Brand\";v=\"99\"",
            "Sec-CH-UA-Platform": "macOS",
            "Sec-CH-UA-Mobile": "?0"
        },
        {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.6; rv:36.0) Gecko/20100101 Firefox/36.0",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.9",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "close",
            "Upgrade-Insecure-Requests": "1",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-User": "?1",
            "Sec-Fetch-Dest": "document",
            "Sec-CH-UA": "\"Chromium\";v=\"120\", \"Microsoft Edge\";v=\"120\", \"Not A;Brand\";v=\"99\"",
            "Sec-CH-UA-Platform": "Windows",
            "Sec-CH-UA-Mobile": "?0"
        }
]

USER_AGENT = approved_headers[3]
S = requests.Session()
S.headers.update({"User-Agent": USER_AGENT})
def get_session(url):
    session = requests.Session()
    session.headers.update(USER_AGENT)
    return session
def fetch_html(url,session = None):
    session = session or get_session(url)
    response = session.get(url)
    return response.text
def get_source_code(url,session = None):
    session = session or get_session(url)
    return fetch_html(url=url,session=session)
def get_soup(url,session = None):
    source_code = get_source_code(url,session = session)
    return BeautifulSoup(source_code, "html.parser")
def get_soup_text(url,session = None):
    source_soup = get_soup(url,session = session)
    return source_soup.text
def get_headers(url,session = None):
    source_soup = get_soup(url,session = session)
def get_all_js(url=None,html=None):
    if not html:
        html = get_html(url)
    all_js = []
    for line in html.split('.js'):
        all_js.append(f"""{line.split('"')[-1]}.js""")
    return all_js



from .imports import *
import threading,os,re,urllib.request,subprocess,requests,shutil,tempfile
import os,re,hashlib,unicodedata
M3U8 = lazy_import('m3u8','M3U8')
m3u8_To_MP4 = lazy_import('m3u8_To_MP4')
yt_dlp= lazy_import('yt_dlp')
PyPDF2= lazy_import('PyPDF2')
FFmpegFixupPostProcessor = lazy_import('yt_dlp','postprocessor','ffmpeg','FFmpegFixupPostProcessor')
logger = get_logFile('video_bp')
def get_output_basename(output_basename=None,output_filename=None,output_ext=None):
    basename = output_basename
    filename = output_filename
    ext = output_ext
    
    if basename:
        if ext and not filename:
            filename,_ = os.path.splitext(basename)
        if filename and not ext:
            _,ext = os.path.splitext(basename)
        if not filename and not ext:
            filename,ext = os.path.splitext(basename)
    if filename:
        if not ext:
            filename,ext = os.path.splitext(filename)
        else:
            filename,_ = os.path.splitext(filename)
    if not filename:
        filename = 'video'
    if not ext:
       ext = '.mp4'
    return f"{filename}{ext}"

class VideoDownloader:
    def __init__(self, url,
                 title=None,
                 download_directory=os.getcwd(),
                 user_agent=None,
                 video_extention=None, 
                 download_video=True,
                 get_info=False,
                 auto_file_gen=True,
                 standalone_download=False,
                 output_filename=None,
                 output_ext=None,
                 output_basename=None,
                 ydl_opts=None,
                 video_id=None,
                 **kwargs
                 ):
        output_basename = get_output_basename(output_basename=output_basename,output_ext=video_extention)
        output_filename,output_ext = os.path.splitext(output_basename)
        
        self.url = url
        self.ydl_opts = ydl_opts or {}
        self.video_id = video_id
        self.monitoring = True
        self.pause_event = threading.Event()
        self.get_download = download_video
        self.get_info = get_info
        self.user_agent = user_agent
        self.title = title
        self.auto_file_gen = auto_file_gen
        self.standalone_download = standalone_download
        self.video_extention = output_ext
        self.output_basename = output_basename
        self.download_directory = download_directory
        self.output_filename = output_filename  # New parameter for custom filename
        self.header = {}  # Placeholder for UserAgentManagerSingleton if needed
        self.outtmpl = os.path.join(self.download_directory,self.output_basename)
        self.video_urls = [self.url]
        self.info = {}
        self.starttime = None
        self.downloaded = 0
        self.video_urls = url if isinstance(url, list) else [url]
        self.send_to_dl()
        


    def send_to_dl(self):
        if self.standalone_download:
            self.standalone_downloader()
        else:
            self.start()

    def get_headers(self, url):
        response = requests.get(url)
        if response.status_code == 200:
            return response.headers
        else:
            logger.error(f"Failed to retrieve headers for {url}. Status code: {response.status_code}")
            return {}

    @staticmethod
    def get_directory_path(directory, name, video_extention):
        file_path = os.path.join(directory, f"{name}.{video_extention}")
        i = 0
        while os.path.exists(file_path):
            file_path = os.path.join(directory, f"{name}_{i}.{video_extention}")
            i += 1
        return file_path

    def progress_callback(self, stream, chunk, bytes_remaining):
        total_size = stream.filesize
        self.downloaded = total_size - bytes_remaining

    def download(self):
        outtmpl = self.outtmpl
        for video_url in self.video_urls:
            # Use custom filename if provided, otherwise generate a short temporary one

            ydl_opts = {
                'external_downloader': 'ffmpeg',
                'outtmpl': self.outtmpl,
                'format': 'bestvideo[ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]/best',
                'merge_output_format': self.video_extention.lstrip('.'),
                'noprogress': True,
                'quiet': True,  # Reduce verbosity in logs
            }
            ydl_opts = {**ydl_opts, **self.ydl_opts}  # merge user opts
            try:
                with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                    self.info = ydl.extract_info(video_url, download=self.get_download)
                    self.downloading = False
                    self.starttime = get_time_stamp()  # Assuming get_time_stamp() exists
                    if self.auto_file_gen:
                        file_path = ydl.prepare_filename(self.info)
                        if self.get_info:
                            self.info['file_path'] = file_path  # Fixed typo 'aath'
                    if self.get_info:
                        self.stop()
                        
            except Exception as e:
                logger.error(f"Failed to download {video_url}: {str(e)}")
            self.stop()
        self.info['video_id']= self.video_id or self.info.get('display_id') or self.info.get('id')
        return self.info

    def monitor(self):
        while self.monitoring:
            logger.info("Monitoring...")
            self.pause_event.wait(60)  # Check every minute
            if self.starttime:
                elapsed_time = subtract_it(get_time_stamp(),self.starttime)
                if self.downloaded != 0 and elapsed_time != 0:
                    cumulative_time = add_it(self.downloaded,elapsed_time)
                    percent = divide_it(self.downloaded,cumulative_time)
                else:
                    percent = 0
                if elapsed_time != 0:
                    try:
                        downloaded_minutes = divide_it(elapsed_time,60)
                        estimated_download_minutes = divide_it(downloaded_minutes,percent)
                        estimated_download_time =  subtract_it(estimated_download_minutes,downloaded_minutes)
                    except ZeroDivisionError:
                        logger.warning("Caught a division by zero in monitor!")
                        continue
                if downloaded_minutes != 0 and subtract_it(percent,downloaded_minutes) != 0:
                    estimated_download_minutes = divide_it(downloaded_minutes,percent)
                    estimated_download_time =  subtract_it(estimated_download_minutes,downloaded_minutes)
                    logger.info(f"Estimated download time: {estimated_download_time} minutes")
                if estimated_download_time >= 1.5:
                    logger.info("Restarting download due to slow speed...")
                    self.start()  # Restart download

    def start(self):
        self.download_thread = threading.Thread(target=self.download)
        self.download_thread.daemon = True
        self.monitor_thread = threading.Thread(target=self.monitor)
        self.download_thread.start()
        self.monitor_thread.start()
        self.download_thread.join()
        self.monitor_thread.join()

    def stop(self):
        self.monitoring = False
        self.pause_event.set()

def download_image(url, save_path=None):
    """
    Downloads an image from a URL and saves it to the specified path.
    
    Args:
        url (str): The URL of the image to download
        save_path (str, optional): Path to save the image. If None, uses the filename from URL
        
    Returns:
        str: Path where the image was saved, or None if download failed
    """
    try:
        # Send GET request to the URL
        response = requests.get(url, stream=True)
        
        # Check if the request was successful
        if response.status_code == 200:
            # Set decode_content=True to automatically handle Content-Encoding
            response.raw.decode_content = True
            
            # If no save_path provided, extract filename from URL
            if save_path is None:
                # Get filename from URL
                filename = url.split('/')[-1]
                save_path = filename
            
            # Ensure the directory exists
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            
            # Write the image content to file
            with open(save_path, 'wb') as f:
                f.write(response.content)
            
            print(f"Image successfully downloaded to {save_path}")
            return save_path
        else:
            print(f"Failed to download image. Status code: {response.status_code}")
            return None
            
    except requests.exceptions.RequestException as e:
        print(f"Error downloading image: {str(e)}")
        return None
    except Exception as e:
        print(f"An unexpected error occurred: {str(e)}")
        return None

def get_thumbnails(directory,info):
    thumbnails_dir = os.path.join(directory,'thumbnails')
    os.makedirs(thumbnails_dir, exist_ok=True)
    thumbnails = info.get('thumbnails',[])
    for i,thumbnail_info in enumerate(thumbnails):
        thumbnail_url = thumbnail_info.get('url')
        thumbnail_base_url = thumbnail_url.split('?')[0]
        baseName = os.path.basename(thumbnail_base_url)
        fileName,ext = os.path.splitext(baseName)
        baseName = f"{fileName}{ext}"
        resolution = info['thumbnails'][i].get('resolution')
        if resolution:
            baseName = f"{resolution}_{baseName}"
        img_id = info['thumbnails'][i].get('id')
        if img_id:
            baseName = f"{img_id}_{baseName}"
        thumbnail_path = os.path.join(thumbnails_dir,baseName)
        info['thumbnails'][i]['path']=thumbnail_path
        download_image(thumbnail_url, save_path=thumbnail_path)
    return info

def optimize_video_for_safari(input_file, reencode=False):
    """
    Optimizes an MP4 file for Safari by moving the 'moov' atom to the beginning.
    Optionally, re-encodes the video for maximum compatibility.
    
    Args:
        input_file (str): Path to the original MP4 file.
        reencode (bool): If True, re-encode the video for Safari compatibility.
        
    Returns:
        str: Path to the optimized MP4 file.
    """
    tmp_dir = tempfile.mkdtemp()
    try:
        local_input = os.path.join(tmp_dir, os.path.basename(input_file))
        shutil.copy2(input_file, local_input)
        
        base, ext = os.path.splitext(local_input)
        local_output = f"{base}_optimized{ext}"
        
        if reencode:
            # Re-encoding command for maximum Safari compatibility
            command = [
                "ffmpeg", "-i", local_input,
                "-c:v", "libx264", "-profile:v", "baseline", "-level", "3.0", "-pix_fmt", "yuv420p",
                "-c:a", "aac", "-b:a", "128k",
                "-movflags", "faststart",
                local_output
            ]
        else:
            # Simple faststart with stream copy
            command = [
                "ffmpeg", "-i", local_input,
                "-c", "copy", "-movflags", "faststart",
                local_output
            ]
        
        try:
            subprocess.run(command, check=True)
            shutil.copy2(local_output, input_file)
            print(f"Optimized video saved as {input_file}")
        except subprocess.CalledProcessError as e:
            print(f"Error during optimization: {e}")
        return input_file
    finally:
        shutil.rmtree(tmp_dir)
def bool_or_default(obj,default=True):
    if obj == None:
        obj =  default
    return obj

def get_temp_id(url):
    url = str(url)
    url_length = len(url)
    len_neg = 20
    len_neg = len_neg if url_length >= len_neg else url_length
    temp_id = re.sub(r'[^\w\d.-]', '_', url)[-len_neg:]
    return temp_id
def get_temp_file_name(url):
    temp_id = get_temp_id(url)
    temp_filename = f"temp_{temp_id}.mp4"
    return temp_filename
def get_display_id(info):
    display_id = info.get('display_id') or info.get('id')
    return display_id
def get_video_title(info):
    title = info.get('title', 'video')[:30]
    return title
def get_safe_title(title):
    re_str = r'[^\w\d.-]'
    safe_title = re.sub(re_str, '_', title)
    return safe_title
def get_video_info_from_mgr(video_mgr):
    try:
        if hasattr(video_mgr,"info"):
            info = video_mgr.info
            return info
    except Exception as e:
        print(f"{e}")
        return None
    return video_mgr
def get_video_mgr(url, download_directory=None, output_filename=None,output_ext=None,output_basename=None,
                   get_info=None, download_video=None, ydl_opts=None,video_id=None,**kwargs):
    directory = download_directory or get_default_videos_dir()
    
    output_basename = get_output_basename(output_basename=output_basename,output_filename=output_filename,output_ext=output_ext)
    output_basename = output_basename or get_temp_file_name(url)
    get_info = bool_or_default(get_info)
    download_video = bool_or_default(download_video, default=False)
    video_mgr = VideoDownloader(
        url=url,
        download_directory=directory,
        download_video=download_video,
        get_info=get_info,
        output_basename=output_basename,
        ydl_opts=ydl_opts,
        video_id=None
        # pass through
    )
    return video_mgr
def downloadvideo(url, download_directory=None, output_filename=None,output_ext=None,output_basename=None,
                   get_info=None, download=True, ydl_opts=None,**kwargs):
    download = download or kwargs.get('download_video')
    output_basename = get_output_basename(output_basename=output_basename,output_filename=output_filename,output_ext=output_ext)
    video_mgr = get_video_mgr(url, directory=directory, output_basename=output_basename,
                   get_info=get_info, download_video=download,ydl_opts=ydl_opts)
    return get_video_info_from_mgr(video_mgr)
def get_video_info(url, download_directory=None, output_filename=None,output_ext=None,output_basename=None,
                   get_info=None, download_video=False, ydl_opts=None,**kwargs):
    output_basename = get_output_basename(output_basename=output_basename,output_filename=output_filename,output_ext=output_ext)
    video_mgr = get_video_mgr(url, download_directory=download_directory, output_basename=output_basename,
                   get_info=get_info, download_video=download_video,ydl_opts=ydl_opts,**kwargs)
    return get_video_info_from_mgr(video_mgr)

def get_video_id(url, download_directory=None, output_filename=None,output_ext=None,output_basename=None,
                   get_info=None, download_video=False, ydl_opts=None,**kwargs):
    output_basename = get_output_basename(output_basename=output_basename,output_filename=output_filename,output_ext=output_ext)
    video_info = get_video_info(url, download_directory=download_directory, output_basename=output_basename,
                   get_info=get_info, download_video=download_video,ydl_opts=ydl_opts,**kwargs)
    return video_info.get('id')
def dl_video(url, download_directory=None, output_filename=None,output_ext=None,output_basename=None,
             get_info=None, download_video=None, ydl_opts=None,**kwargs):
    output_basename = get_output_basename(output_basename=output_basename,output_filename=output_filename,output_ext=output_ext)
    mgr = get_video_info(
        url,
        download_directory=download_directory,
        output_basename=output_basename,
        get_info=get_info,
        download_video=download_video,
        ydl_opts=ydl_opts,**kwargs)
    return get_video_info_from_mgr(mgr)
def for_dl_video(url,download_directory=None,output_filename=None,get_info=None,download_video=None,output_ext=None,output_basename=None,**kwargs):
    get_info = bool_or_default(get_info,default=True)
    download_video =bool_or_default(download_video,default=True)
    output_basename = get_output_basename(output_basename=output_basename,output_filename=output_filename,output_ext=output_ext)
    video_mgr = dl_video(url,download_directory=download_directory,output_basename=output_basename,get_info=get_info,download_video=download_video,**kwargs)
    if get_video_info_from_mgr(video_mgr):
        return get_video_info_from_mgr(video_mgr)
    videos = soupManager(url).soup.find_all('video')
    for video in videos:
        src = video.get("src")
        video_mgr = dl_video(src,download_directory=download_directory,output_basename=output_basename,get_info=get_info,download_video=download_video,**kwargs)
        if get_video_info_from_mgr(video_mgr):
            return get_video_info_from_mgr(video_mgr)

download_video = downloadvideo
downloadVideo = downloadvideo



TENOR_HOSTS = {"tenor.com", "www.tenor.com"}

def detect_tenor(parsed):
    if parsed.netloc.lower() not in TENOR_HOSTS:
        return None

    parts = parsed.path.split("/")
    if len(parts) > 2 and parts[-1].isdigit():
        return {"downloadable": True, "kind": "image", "provider": "tenor",
                "id": parts[-1], "direct": False}

    return {"downloadable": False, "provider": "tenor"}



IMGUR_HOSTS = {"imgur.com", "i.imgur.com"}

def detect_imgur(parsed):
    net = parsed.netloc.lower()
    if net not in IMGUR_HOSTS:
        return None

    ext = os.path.splitext(parsed.path)[1].lower()
    if ext:
        return {"downloadable": True, "kind": "image", "provider": "imgur",
                "id": parsed.path.strip("/"), "direct": True}

    return {"downloadable": False, "provider": "imgur"}



from .manager_utils import *
from .playwriteManager import *
from .cipherManager import *
from .clownworld import *
from .crawlManager import *
from .imitationBrowser import *
from .linkManager import *
from .networkManager import *
from .playwriteManager import *
from .requestManager import *
from .seleneumManager import *
from .soupManager import *
from .sslManager import *
from .tlsAdapter import *
from .urlManager import *
from .userAgentManager import *

from .usurpManager import *
from .manager_utils import *
from .videoDownloader import *
##from .middleManager import get_soup_tools,get_req_tools,get_url_tools
##
##seleniumManager = seleneumManager
##
